unit Unit1;
{ Lecteur de fichier GRAPPA.

  PREND EN COMPTE LES DONNES UCAC4 ISSUES DE LA FUSION DE LA BASE D'OCCULT 

  Il permet de lire un fichier GRAPPAE3 et d'en extraire toutes les infos sur une source

  2013-2021 Marc SERRAU

  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA.
}
{$mode objfpc}{$H+}

interface

uses
   Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
   StdCtrls, Buttons, LCLType, Clipbrd, recherche, Unit2, cu_healpix;

type

   { TForm1 }

   TForm1 = class(TForm)
     ButtonSearchFile: TButton;
      ButtonRecherche: TButton;
      CheckBoxUCAC4: TCheckBox;
      CheckBoxDuplicated: TCheckBox;
      CheckBoxHasBP: TCheckBox;
      CheckBoxHasRP: TCheckBox;
      CheckBoxHasHipparcosID: TCheckBox;
      CheckBoxHasTychoID: TCheckBox;
      CheckBoxHasRV: TCheckBox;
      CheckBoxHasPM_par_RUWE: TCheckBox;
      CheckBoxHasG: TCheckBox;
      EditNumRec: TEdit;
      EditUCAC4: TEdit;
      EditSourceIDRecherche: TEdit;
      EditBP: TEdit;
      EditRP: TEdit;
      EditHipparcosID: TEdit;
      EditG: TEdit;
      EditTychoID: TEdit;
      Label1: TLabel;
      Label2: TLabel;
      Label3: TLabel;
      Label4: TLabel;
      Label5: TLabel;
      Label6: TLabel;
      LabeledEditRAin: TLabeledEdit;
      LabeledEditPMDEC: TLabeledEdit;
      LabeledEditParallax: TLabeledEdit;
      LabeledEditDECin: TLabeledEdit;
      LabeledEditRUWE: TLabeledEdit;
      LabeledEditRV: TLabeledEdit;
      LabeledEditPMRA: TLabeledEdit;
      LabelSourceID: TLabel;
      LabeledEditAD: TLabeledEdit;
      LabeledEditDEC: TLabeledEdit;
      LabelNombre: TLabel;
      LabeledEditNomFichier: TLabeledEdit;
      OpenDialog1: TOpenDialog;
      SelectionRepBase: TSelectDirectoryDialog;
      SpeedButtonDown: TSpeedButton;
      SpeedButtonUp: TSpeedButton;
      procedure ButtonRechercheClick(Sender: TObject);
      procedure ButtonSearchFileClick(Sender: TObject);
      procedure EditNumRecChange(Sender: TObject);
      procedure FormActivate(Sender: TObject);
      procedure FormCreate(Sender: TObject);
      procedure FormDestroy(Sender: TObject);
      procedure LabeledEditNomFichierClick(Sender: TObject);
      procedure AfficheGrappa(p : longint);
      procedure LabelSourceIDClick(Sender: TObject);
      procedure SpeedButtonDownClick(Sender: TObject);
      procedure SpeedButtonUpClick(Sender: TObject);

   private
      { private declarations }
   public
      { public declarations }
   end;

const
   BASE_defaut = 'I:\'; // Where to begin to search the GRAPPA-E3 database (for asking)
   // All these files must be present in the intallation directory:
   GAIA_EDR3_Crossmatch_Hipparcos = 'Hipparcos2BestNeighbour.dat';
   GAIA_EDR3_Crossmatch_Tycho     = 'Tycho2tdscMergeBestNeighbour.dat';
   BASE_VR                        = 'VR.dat';
   BASE_UCAC                      = 'UCAC4_crossmatch.dat';

type
   TGRAPPA = bitpacked record
                           //         Size
                           // Unit   (byte) Definition
      Source_ID : QWord;   //          8    Unique within a particular Data Release. See https://dms.cosmos.esa.int/COSMOS/doc_fetch.php?id=2779219
      RA        : int32;   // mas      4    R.A.truncated to 1mas precision truncation: +10.123 = 10
      RA2       : byte;    // 4µas     1    R.A.component <1mas. Units of 4 µas. To be added to RA. 30 = 0.120mas
      e_RA      : uint16;  // 10µas    2    Uncertainty in RA units of 10 µas
      DEC       : int32;   // mas      4    Declination+90° truncated 1 mas precision truncation: +10.123 = 10
      DEC2      : byte;    // 4µas     1    DEC component <1mas. Units of 4 µas. To be added to DEC. 30 = 0.120mas
      e_DEC     : uint16;  // 10µas    2    Uncertainty in Dec, units of 10 µas
      parallax  : uint16;  // 12.5µas  2    Permet de coder la parallax de Proxima Cen.
      e_parallax: uint16;  // 10µas    2    Uncertainty in Parallax units of 10 µas
      PM_RA     : int32;   // µas/yr   4    Annual proper motion in RA. Value has been multiplied by Cos(declination)
      e_PM_RA   : uint16;  // µas/yr   2    Uncertainty in Proper motion in RA
      PM_DEC    : int32;   // µas/yr   4    Annual proper motion in Dec
      e_PM_DEC  : uint16;  // µas/yr   2    Uncertainty in Proper motion in Dec
      RUWE      : byte;    //   0.1    1    Value for ruwe indicator See : https://dms.cosmos.esa.int/COSMOS/doc_fetch.php?id=3757412
                           //               RUWE is set to 0 for sources with only two parameters (RA & DEC)
      has_RV,              //         1/8   Set if there is a radial velocity for the star in the RV.dat file
      has_PM_parallax_RUWE,//         1/8   Set if there are PM, parallax and RUWE values for the star
      has_hipparcos2_id,   //         1/8   Set if an Hipparcos2 crossmatch has been found
      has_TYCHO2DSC_id,    //         1/8   Set if an Tycho2DSC crossmatch has been found
      has_G,has_BP,has_RP, //       3x1/8   Set if magG, magBP, magRP are given respectively
      isDuplicatedSource : boolean;// 1/8   Set if this source happened to been duplicated and one source only has been kept
      G,BP,RP   : int32;   // 0.1mmag 3x4   Values for magnitude in the G, BP and RP bands respectively including e_G, e_BP, e_RP
                           //               Coded as : 18bits for value (value can go up to 26.2144), 14 bits for incertitudes.
   //      e_G,e_BP,e_RP : byte;// 0.1mmag       Incertitudes on G, BP and RP in 0.1mmag
   end;

   TCrossmatch_index = packed record
      source_id : QWord;
      original_ext_source_id : packed array[1..11] of char;
      angular_distance : single;

   end;

   TTycho_index = packed record
      source_id : QWord;
      original_ext_source_id : packed array[1..11] of char;
      angular_distance : single;
      number_of_neighbours,xm_flag : byte;
      tycho2tdsc_merge_oid : longint;
   end;

   THipparcos_index = packed record
      source_id : QWord;
      original_ext_source_id : longint;
      angular_distance : single;
      number_of_neighbours,xm_flag : byte;
   end;
                         //        Size
   TVR = packed record   //  Unit (byte)  Definition
      source_id : QWord; //    -     8
      VR        : smallint;//km/s    2    Radial velocity    (Smallint = -32768..32767)
      e_VR      : byte;  //  km/s    1    Uncertainty in Radial Velocity
   end;

   TGrappaUCACIndex = packed record
      Source_ID : QWord;   //        8    Unique within a particular Data Release.
      CatNum    : uint32;  //        4    Sequential number in catalogue zone (or number in catalogue)
      match     : byte;    //        1    +1 if UCAC4 identifier based on a poor match
                           //             +2 if Gaia does not contain a proper motion for the star, with the UCAC proper motion using the UCAC4 position at 2000.0
   end;
var
   Form1: TForm1;
   Base : AnsiString;
   env_ok : boolean;
   f_hipparcos_xmatch : file of THipparcos_index;
   f_tycho_xmatch     : file of TTycho_index;

   buffer_hipparcos   : THipparcos_index;
   buffer_tycho       : TTycho_index;

   Index_Hipparcos,
   Index_Tycho,
   Index_VR           : P_Table;

   f_VR               : file of TVR;
   buffer_VR          : TVR;
   fUCAC              : file of TGrappaUCACIndex;
   bUCAC              : TGrappaUCACIndex;
   tUCAC              : Longint;

   fichier_ouvert : boolean;
   fgrappa : file of TGRAPPA;
   rec     : TGRAPPA;
   lfGRAPPA,
   fpos,iHipparcos,iTycho,iVR,nb_hipparcos_xmatch,nb_tycho_xmatch,nb_VR : longint;


implementation

{$R *.lfm}

{ TForm1 }
procedure TForm1.FormActivate(Sender: TObject);

begin
end;
procedure TForm1.FormCreate(Sender: TObject);
var
   i : Longint;
begin
   // Lecture des index crossmatch TychoTSC et Hipparchos
   SelectionRepBase.InitialDir:=BASE_defaut;
   SelectionRepBase.Title:='Select your GRAPPAE3 installation directory';

   {GAIA_EDR3_Crossmatch_Hipparcos...   }
   if SelectionRepBase.Execute then
      Base:=SelectionRepBase.FileName+'\';

   env_ok:=false;
   if not FileExists(base+GAIA_EDR3_Crossmatch_Hipparcos) or
      not FileExists(base+GAIA_EDR3_Crossmatch_Tycho) or
      not FileExists(base+BASE_VR) or
      not FileExists(base+BASE_UCAC) then
   begin
     Application.MessageBox('Fail to find one of the index files','Error');
     Application.Terminate;
   end
   else
   begin
     Application.CreateForm(TForm2, Form2);
     Form2.Show;
     Form2.Update;
     Assignfile(f_hipparcos_xmatch,base+GAIA_EDR3_Crossmatch_Hipparcos);
     Reset(f_hipparcos_xmatch);
     nb_hipparcos_xmatch:=FileSize(f_hipparcos_xmatch);
     New(Index_Hipparcos);
     SetLength(Index_Hipparcos^,nb_hipparcos_xmatch+1);
     for i:=1 to nb_hipparcos_xmatch do
     begin
       Read(f_hipparcos_xmatch,buffer_hipparcos);
       New(Index_Hipparcos^[i]);
       With Index_Hipparcos^[i]^ do
       begin
          Index_Hipparcos^[i]^.rang:=i;
          Index_Hipparcos^[i]^.value:=buffer_hipparcos.source_id;
       end;
     end;

     {GAIA_EDR3_Crossmatch_Tycho2DSC...    }

     Assignfile(f_Tycho_xmatch,base+GAIA_EDR3_Crossmatch_Tycho);
     Reset(f_Tycho_xmatch);
     nb_Tycho_xmatch:=FileSize(f_Tycho_xmatch);
     New(Index_Tycho);
     SetLength(Index_Tycho^,nb_Tycho_xmatch+1);
     for i:=1 to nb_Tycho_xmatch do
     begin
       Read(f_Tycho_xmatch,buffer_Tycho);
       New(Index_Tycho^[i]);
       With Index_Tycho^[i]^ do
       begin
          Index_Tycho^[i]^.rang:=i;
          Index_Tycho^[i]^.value:=buffer_Tycho.source_id;
       end;
     end;
  // On ne referme pas le fichier GAIA_EDR3_Crossmatch_Tycho2DSC car on en aura besoin plus tard
  //
  // Lecture du fichier de VR et création de l'index VR en mémoire
  //
     Assignfile(f_VR,base+BASE_VR);
     Reset(f_VR);
     nb_VR:=FileSize(f_VR);
     New(Index_VR);
     SetLength(Index_VR^,nb_VR+1);
     for i:=1 to nb_VR do
     begin
       Read(f_VR,buffer_VR);
       New(Index_VR^[i]);
       With Index_VR^[i]^ do
       begin
          Index_VR^[i]^.rang:=i;
          Index_VR^[i]^.value:=buffer_VR.source_id;
       end;
     end;
     LabeledEditNomFichier.Text:=Base;
  //
  // Ouverture du fichier de crossmatch avec UCAC4 :

     Assignfile(fUCAC,base+BASE_UCAC);
     Reset(fUCAC);
     tUCAC:=Filesize(fUCAC);
     Form2.Hide;

     env_ok:=true;
     LoadHealpixLibrary;
   end;
end;

function RechercheIndexU(rech : QWord):longint;
var
   lim,i,anc,inf,sup     : longint;
   fin,trouve : boolean;
   buffer     : TGrappaUCACIndex;
begin
   // Le fichier fUCAC est ouvert au préalable :

   // Petite recherche dichotomique :

   lim:=trunc(ln(tUCAC)/ln(2));
   trouve:=false;
   fin:=false;
   anc:=trunc(tUCAC/2);
   inf:=0;
   i:=0;
   sup:=tUCAC;
   repeat
       if i>(lim+1) then fin:=true;
       Seek(fUCAC,anc);
       read(fUCAC,buffer);
       if buffer.Source_ID=rech then
          trouve:=true
       else
       begin
          inc(i);
          if buffer.Source_ID > rech then
          begin
             sup:=anc;
             anc:=trunc((inf+anc)/2);
          end
          else
          begin
             inf:=anc;
             anc:=trunc((sup+anc)/2);
          end
       end
   until trouve or fin;
   if trouve then RechercheIndexU:=anc else RechercheIndexU:=-1;
end;

procedure TForm1.AfficheGrappa(p : Longint);
var
   s : string;
   v : extended;
   NumUCAC : Longint;

begin
   Seek(fGRAPPA,p);
   Read(fGRAPPA,rec);
   with rec do
   begin
      Str(p,s);
      EditNumRec.Text:=s;
      Str(Source_ID,s);
      LabelSourceID.Caption:=s;
      v:=(RA+RA2/250)/3600000;
      Str(V:22:18,s); LabeledEditAD.Caption:=s;
      v:=(DEC+DEC2/250)/3600000-90;
      Str(V:22:18,s); LabeledEditDEC.Caption:=s;

      if has_PM_parallax_RUWE then begin
         CheckBoxHasPM_par_RUWE.checked:=true;
         v:=parallax*0.0125;
         Str(V:9:4,s); LabeledEditParallax.Caption:=s;
         v:=e_parallax/100;
         Str(V:9:4,s); s:=Trim(s); LabeledEditParallax.Caption:=LabeledEditParallax.Caption+'+/-'+s;
         V:=ruwe*0.1;
         Str(V:4:2,s); LabeledEditRUWE.Caption:=s;
         V:=PM_RA/1000;
         Str(V:9:4,s); LabeledEditPMRA.Caption:=s;
         v:=e_PM_RA/1000;
         Str(V:9:4,s); s:=Trim(s); LabeledEditPMRA.Caption:=LabeledEditPMRA.Caption+'+/-'+s;
         V:=PM_DEC/1000;
         Str(V:9:4,s); LabeledEditPMDEC.Caption:=s;
         v:=e_PM_DEC/1000;
         Str(V:9:4,s); s:=Trim(s); LabeledEditPMDEC.Caption:=LabeledEditPMDEC.Caption+'+/-'+s;
      end
      else begin
         CheckBoxHasPM_par_RUWE.checked:=false;
         LabeledEditParallax.Caption:='';
         LabeledEditPMRA.Caption:='';
         LabeledEditPMDEC.Caption:='';
         LabeledEditRUWE.Caption:='';
      end;

      if has_RV then begin
         CheckBoxHasRV.Checked:=true;
         iVR:=RechercheIndex(Index_VR,nb_VR,source_id);
         if iVR>0 then
         begin
            Seek(f_VR,iVR-1);          // on cherche l'enregistrement dans de fichier d'index
            Read(f_VR,buffer_VR);      // On le lit
            Str(buffer_VR.VR,s); s:=Trim(s);
            LabeledEditRV.Text:=s;
            Str(buffer_VR.e_VR,s);
            LabeledEditRV.Text:=LabeledEditRV.Text+'+/-'+s;
            if rec.Source_ID=buffer_VR.source_id then
               LabeledEditRV.Text:=LabeledEditRV.Text+' Checked'; // Vérification de l'égalité des source_id

         end;

      end else
      begin
         CheckBoxHasRV.Checked:=false;
         LabeledEditRV.Text:='';
      end;

      if isDuplicatedSource then CheckBoxDuplicated.Checked:=true else CheckBoxDuplicated.Checked:=false;

      if has_Hipparcos2_id then
      begin
         CheckBoxHasHipparcosID.Checked:=true;
         iHipparcos:=RechercheIndex(Index_Hipparcos,nb_hipparcos_xmatch,source_id);
         if iHipparcos>0 then
         begin
            Seek(f_hipparcos_xmatch,iHipparcos-1);          // on cherche l'enregistrement dans de fichier d'index
            Read(f_hipparcos_xmatch,buffer_hipparcos);      // On le lit
            Str(buffer_hipparcos.original_ext_source_id,s); // conversion de l'ID Hipparcos
            if rec.Source_ID=buffer_hipparcos.source_id then s:=s+' Checked'; // Vérification de l'égalité des source_id
            EditHipparcosID.Text:=s;
         end;
      end
      else begin
         CheckBoxHasHipparcosID.Checked:=false;
         EditHipparcosID.Text:='';
      end;

      if has_TYCHO2DSC_id then
      begin
         CheckBoxHasTychoID.Checked:=true;
         iTycho:=RechercheIndex(Index_Tycho,nb_Tycho_xmatch,source_id);
         if iTycho>0 then
         begin
            Seek(f_Tycho_xmatch,iTycho-1);          // on cherche l'enregistrement dans de fichier d'index
            Read(f_Tycho_xmatch,buffer_Tycho);      // On le lit
            s:=Trim(buffer_Tycho.original_ext_source_id); // conversion de l'ID Hipparcos
            if rec.Source_ID=buffer_Tycho.source_id then s:=s+' Checked'; // Vérification de l'égalité des source_id
            EditTychoID.Text:=s;
         end;
      end else begin
        EditTychoID.Text:='';
        CheckBoxHasTychoID.Checked:=false;
      end;

      if has_G then begin
         CheckBoxHasG.checked:=true;
         v:=(G AND $3FFFF)/10000;
         Str(v:9:4,s); EditG.Text:=s;
         v:=((G AND $FFFC0000) shr 18)/10000;
         Str(v:6:4,s); EditG.Text:=EditG.Text+'+/-'+s;
      end else begin
        EditG.Text:='';
        CheckBoxHasG.checked:=false;
      end;

      if has_BP then
      begin
         CheckBoxHasBP.checked:=true;
         v:=(BP AND $3FFFF)/10000;
         Str(v:9:4,s); EditBP.Text:=s;
         v:=((BP AND $FFFC0000) shr 18)/10000;
         Str(v:6:4,s); EditBP.Text:=EditBP.Text+'+/-'+s;
      end else begin
         EditBP.Text:='';
         CheckBoxHasBP.checked:=false;
      end;

      if has_RP then
      begin
         CheckBoxHasRP.checked:=true;
         v:=(RP AND $3FFFF)/10000;
         Str(v:9:4,s); EditRP.Text:=s;
         v:=((RP AND $FFFC0000) shr 18)/10000;
         Str(v:6:4,s); EditRP.Text:=EditRP.Text+'+/-'+s;
      end else begin
         EditRP.Text:='';
         CheckBoxHasRP.checked:=false;
      end;

      // Recherche systématique de la présence de source_id dansle fichier de
      // crossmatch avec index UCAC4:

      NumUCAC:=RechercheIndexU(source_id);
      if NumUCAC>=0 then
      begin
         Seek(fUCAC,NumUCAC); read(fUCAC,bUCAC);
         Str(bUCAC.CatNum,s);
         s:=Copy('000',1,9-Length(s))+s;
         s:=Copy(s,1,3)+'-'+Copy(s,4,6);
         EditUCAC4.Text:=s;
         if bUCAC.match>0 then
            if bUCAC.match>1 then s:='match with PM from UCAC4'
            else
               s:='Poor match'
         else
            s:='Good match';

         EditUCAC4.Text:=EditUCAC4.Text+' '+s;
         CheckBoxUCAC4.Checked:=true;
      end else begin
         EditUCAC4.Text:='';
         CheckBoxUCAC4.checked:=false;
      end;

   end;
end;

procedure TForm1.LabelSourceIDClick(Sender: TObject);
begin
   Clipboard.AsText:=LabelSourceID.Caption;
end;


procedure TForm1.EditNumRecChange(Sender: TObject);
var
    code : Integer;
begin
   Val(EditNumRec.Text,fpos,code);
   if code=0 then
      if (fpos>=0) and (fpos<=lfGRAPPA) then AfficheGrappa(fpos);
end;


procedure TForm1.ButtonRechercheClick(Sender: TObject);
var
    code,indexnumber,
    trouve              : longint;
    SourceRecherche     : QWord;

    NSide               : int64;
    iRA,iDEC            : integer;
    theta,phi,RA,Decl   : double;
    nf,oldf             : AnsiString;
    ssize,sRA,sDEC      : string;
    cas                 : 1..10;

    function cherche_source_id(SourceRecherche : QWord;iRA,iDEC : integer): longint;
    var
       BufferG    : TGrappa;
       i,resultat : longint;
    begin
       resultat:= -1;
       if iDEC<5 then nf:=Base+'S85-90'  // Cas de la calotte Sud
       else
          if iDEC>174 then               // Cas de la calotte Nord
             nf:=Base+'N85-90'
          else
          begin                            // Cas général (à 0,7' près)
             Str(iRA,sRA);
             Str(iDEC,sDEC);
             nf:=Base+sDEC+'\'+sRA+'-'+sDEC;
          end;

       if FileExists(nf) then   // on teste quand car la base peut être incomplète
       begin
          LabeledEditNomFichier.Text:=nf;
          If fichier_ouvert then Closefile(fGRAPPA);

          fichier_ouvert:=false;
          Assignfile(fGRAPPA,LabeledEditNomFichier.Text);
          Reset(fGRAPPA);
          fichier_ouvert:=true;

          lfGRAPPA:=Filesize(fGRAPPA)-1;
          Str(lfGRAPPA,ssize);
          LabelNombre.Caption:=' de 0 à '+SSize;
          EditNumRec.Caption:='0';
          fpos:=0;

          i:=-1;
          repeat
             inc(i);
             Seek(fGRAPPA,i);
             Read(fGRAPPA,bufferG);

          until (bufferG.Source_ID=SourceRecherche) or (i=lfGRAPPA);
          if bufferG.Source_ID=SourceRecherche then
             resultat:=i
          else
          begin
             Fichier_ouvert:=false;
             Closefile(fGRAPPA);
          end;
       end;
       cherche_source_id:=resultat;
    end;

begin

   Val(EditSourceIDRecherche.Text,SourceRecherche,code);
   if code=0 then      // On a bien donné un nombre
   begin
      oldf:=LabeledEditNomFichier.Text;

      IndexNumber:=(SourceRecherche AND $FFFFFFF800000000) SHR 35;
      NSide:=4096; {201326592;}
      phi:=0; Theta:=0;
// find coordinates from pixel
      pix2ang_nest64(NSide,IndexNumber,theta,phi);
      // on a obtenu les coordonnées Theta, phi du centre du healpixel.
      // A 0,7' près cela donne les coordonnées de la source donc on peut
      // commencer à chercher le fichier à partir du RA, DEC correspondant :

      RA:=180*phi/pi;
      Decl:=90-180*theta/pi;

      iRA:=Trunc(RA);
      iDEC:=Trunc(90+Decl);  // iDEC = indice <> Déclinaison

      trouve:=-1;
      cas:=1;
      repeat   // Boucle de recherche bête et méchante :
         case cas of
         1 : trouve:=cherche_source_id(SourceRecherche,iRA,iDEC);
         2 : trouve:=cherche_source_id(SourceRecherche,(iRA-1)mod 360,iDEC); { Plus à l'Ouest}
         3 : trouve:=cherche_source_id(SourceRecherche,(iRA+1)mod 360,iDEC); { Pls à l'Est }
         4 : trouve:=cherche_source_id(SourceRecherche,iRA,iDEC+1); { Plus au Nord }
         5 : trouve:=cherche_source_id(SourceRecherche,iRA,iDEC-1); { Plus au Sud }
         6 : trouve:=cherche_source_id(SourceRecherche,iRA-1,iDEC+1); { Plus au Nord-Ouest }
         7 : trouve:=cherche_source_id(SourceRecherche,iRA-1,iDEC-1); { Plus au Sud-Ouest }
         8 : trouve:=cherche_source_id(SourceRecherche,iRA+1,iDEC+1); { Plus au Sud-Est }
         9 : trouve:=cherche_source_id(SourceRecherche,iRA+1,iDEC-1); { Plus au Sud-Est }
         end;
         Inc(cas)
      until (cas=10) or (trouve>=0);
      Dec(Cas);

      if trouve>=0 then
      begin
         fpos:=trouve;
//            EditNumRec.Enabled:=true;
         SpeedButtonDown.Enabled:=true;
         SpeedButtonUp.Enabled:=true;
         EditSourceIDRecherche.Enabled:=true;
//            ButtonRecherche.Enabled:=true;
         AfficheGrappa(fpos);
      end
      else
         LabeledEditNomFichier.Text:=oldf;

   end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
    i: longint;
begin
   if fichier_ouvert then Closefile(fGRAPPA);
   if env_ok then  // L'application avait pu démarrer correctement
   begin
     Closefile(f_Hipparcos_xmatch);
     Closefile(f_Tycho_xmatch);
     Closefile(f_VR);
     Closefile(fUCAC);

  // Libération de la mémoire des index_tycho et index_hipparcos :
     for i:=1 to nb_tycho_xmatch do
        Dispose(Index_Tycho^[i]);
     SetLength(Index_Tycho^,0);
     Dispose(Index_Tycho);
  //   Writeln(esp,DateTimeToStr(Now),' Index Tycho2DSC effacé de la mémoire');

     for i:=1 to nb_hipparcos_xmatch do
        Dispose(Index_hipparcos^[i]);
     SetLength(Index_hipparcos^,0);
     Dispose(Index_hipparcos);


  // On efface l'index VR :
     for i:=1 to nb_VR do
        Dispose(Index_VR^[i]);
     SetLength(Index_VR^,0);
     Dispose(Index_VR);
   end

end;

procedure TForm1.SpeedButtonDownClick(Sender: TObject);
begin
   Dec(fpos);
   if fpos<0 then fpos:=0
   else
      AfficheGrappa(fpos);
end;

procedure TForm1.SpeedButtonUpClick(Sender: TObject);
begin
   Inc(fpos);
   if fpos>lfGrappa then fpos:=lfGrappa
   else
      AfficheGrappa(fpos);
end;

procedure TForm1.LabeledEditNomFichierClick(Sender: TObject);
var
   nf : ansistring;
   ssize : string;
begin
   OpenDialog1.Filter:='Fichier GRAPPA3E|*.*';
   OpenDialog1.InitialDir:=ExtractFilePath(LabeledEditNomFichier.Text);
   if OpenDialog1.execute then
   Begin
      nf:=OpenDialog1.FileName;
      If FileExists(nf) and (Filesize(nf)<>0) then
      begin
         LabeledEditNomFichier.Text:=nf;

         If fichier_ouvert then
         begin
            Closefile(fGRAPPA);
          end;
         fichier_ouvert:=false;
         Assignfile(fGRAPPA,LabeledEditNomFichier.Text);
         Reset(fGRAPPA);
         lfGRAPPA:=Filesize(fGRAPPA)-1;
         Str(lfGRAPPA,ssize);
         LabelNombre.Caption:=' de 0 à '+SSize;

         fpos:=0;

         fichier_ouvert:=true;
         EditNumRec.Enabled:=true;
         SpeedButtonDown.Enabled:=true;
         SpeedButtonUp.Enabled:=true;
         ButtonRecherche.Enabled:=true;
         EditSourceIDRecherche.Enabled:=true;

         AfficheGrappa(fpos);

      end;
   end;
end;
procedure TForm1.ButtonSearchFileClick(Sender: TObject);
var
   RAin , DECin : real;
   iRAin, iDECin,code : integer;
   nf : ansistring;
   ssize,sRAin,sDECin : string;
begin
   Val(LabeledEditRAin.Text,RAin,code);
   if (code=0) and (RAin>=0) and (RAin<360) then
   begin
      Val(LabeledEditDECin.Text,DECin,code);
      if (code=0) and (DECin>=-90) and (DECin<90) then
      begin
         iRAin:=Trunc(RAin);
         iDECin:=Trunc(DECin+90);
         Str(iRAin,sRAin);

         if iDECin<5 then nf:=Base+'S85-90'  // Cas de la calotte Sud
         else
            if iDECin>174 then               // Cas de la calotte Nord
               nf:=Base+'N85-90'
            else
            begin                            // Cas général
               Str(iDECin,sDECin);
               nf:=Base+sDECin+'\'+sRAin+'-'+sDECin;
            end;
         if FileExists(nf) then   // on teste quand car la base peut être incomplète
         begin
            LabeledEditNomFichier.Text:=nf;
            fichier_ouvert:=false;
            Assignfile(fGRAPPA,LabeledEditNomFichier.Text);
            If fichier_ouvert then Closefile(fGRAPPA);
            Reset(fGRAPPA);
            lfGRAPPA:=Filesize(fGRAPPA)-1;
            Str(lfGRAPPA,ssize);
            LabelNombre.Caption:=' de 0 à '+SSize;
            EditNumRec.Caption:='0';
            fpos:=0;

            fichier_ouvert:=true;
            EditNumRec.Enabled:=true;
            SpeedButtonDown.Enabled:=true;
            SpeedButtonUp.Enabled:=true;
            EditSourceIDRecherche.Enabled:=true;
            ButtonRecherche.Enabled:=true;

            AfficheGrappa(fpos);

         end
      end
   end
end;

end.

